import os
import re
import math
import shlex
from shutil import which
from subprocess import PIPE, DEVNULL, Popen, check_output
from typing import Optional, Union

from .exceptions import (
    FFmpegError,
    FFmpegFailedToExtractFrames,
    FFmpegNotFound,
    FramesExtractorOutPutDirDoesNotExist,
)
from .utils import does_path_exists

# python module to extract the frames from the input video.
# Uses the FFmpeg Software to extract the frames.


class FramesExtractor:
    """
    Extract frames from the input video file and save at the output directory(frame storage directory).
    """

    def __init__(
        self,
        video_path: str,
        output_dir: str,
        video_length: float,
        interval: Union[int, float] = 1,
        ffmpeg_path: Optional[str] = None,
    ) -> None:
        """
        Raises Exeception if video_path does not exists.
        Raises Exeception if output_dir does not exists or if not a directory.

        Checks  the ffmpeg installation and the path; thus ensure that we can use it.

        :return: None

        :rtype: NoneType

        :param video_path: absolute path of the video

        :param output_dir: absolute path of the directory
                           where to save the frames.

        :param interval: interval is seconds. interval must be an integer.
                         Extract one frame every given number of seconds.
                         Default is 1, that is one frame every second.

        :param ffmpeg_path: path of the ffmpeg software if not in path.

        """
        self.video_path = video_path
        self.output_dir = output_dir
        self.video_length = video_length
        self.interval = interval
        self.ffmpeg_path = ""
        if ffmpeg_path:
            self.ffmpeg_path = ffmpeg_path

        if not does_path_exists(self.video_path):
            raise FileNotFoundError(
                f"No video found at '{self.video_path}' for frame extraction."
            )

        if not does_path_exists(self.output_dir):
            raise FramesExtractorOutPutDirDoesNotExist(
                f"No directory called '{self.output_dir}' found for storing the frames."
            )

        self._check_ffmpeg()

        self.extract()

    def _check_ffmpeg(self) -> None:
        """
        Checks the ffmpeg path and runs 'ffmpeg -version' to verify that the
        software, ffmpeg is found and works.

        :return: None

        :rtype: NoneType
        """

        if not self.ffmpeg_path:

            if not which("ffmpeg"):

                raise FFmpegNotFound(
                    "FFmpeg is not on the system path. Install FFmpeg and add it to the path."
                    + "Or you can also pass the path via the 'ffmpeg_path' parameter."
                )
            else:

                self.ffmpeg_path = str(which("ffmpeg"))

        # Check the ffmpeg
        try:
            # check_output will raise FileNotFoundError if it does not finds the ffmpeg
            output = check_output([str(self.ffmpeg_path), "-version"]).decode()

        except FileNotFoundError:
            raise FFmpegNotFound(f"FFmpeg not found at '{self.ffmpeg_path}'.")

        else:

            if "ffmpeg version" not in output:
                raise FFmpegError(
                    f"ffmpeg at '{self.ffmpeg_path}' is not really ffmpeg. Output of ffmpeg -version is \n'{output}'."
                )

    @staticmethod
    def detect_crop(
        video_path: Optional[str] = None,
        frames: int = 3,
        ffmpeg_path: Optional[str] = None,
        video_length: float = 2
    ) -> list:
        """
        Detects the the amount of cropping to remove black bars.

        The method uses [ffmpeg.git] / libavfilter /vf_cropdetect.c
        to detect_crop for some fixed intervals.

        The mode of the detected crops is selected as the crop required.

        :return: FFmpeg argument -vf filter and confromable crop parameter.

        :rtype: str
        """

        # We look upto the 120th minute into the video to detect the most
        # precise crop value
        time_start_list = [
            2,
            5,
            10,
            20,
            40,
            100,
            300,
            600,
            1200,
            2400,
            7200,
            14400,
        ]
        

        crop_list = []

        for start_time in time_start_list:

            # Stopping the loop if we go beyond the end length of the video.
            # We round the video length up to make sure we do get the whole
            # video.
            if start_time > math.ceil(video_length):
                break

            command = f'"{ffmpeg_path}" -ss {start_time} -i "{video_path}" -vframes {frames} -vf cropdetect -f null -'

            process = Popen(shlex.split(command), stdin=DEVNULL, stdout=PIPE, stderr=PIPE)

            output, error = process.communicate()

            matches = re.findall(
                r"crop\=[1-9][0-9]{0,3}:[1-9][0-9]{0,3}:[0-9]{1,4}:[0-9]{1,4}",
                (output.decode() + error.decode()),
            )

            for match in matches:
                crop_list.append(match)

        mode = None
        if len(crop_list) > 0:
            mode = max(crop_list, key=crop_list.count)

        crop = []
        if mode:
            crop = ["-vf", mode]

        return crop

    def extract(self) -> None:
        """
        Extract the frames at every n seconds where n is the
        integer set to self.interval.

        :return: None

        :rtype: NoneType
        """

        ffmpeg_path = self.ffmpeg_path
        video_path = self.video_path
        video_length = self.video_length
        output_dir = self.output_dir

        if os.name == "posix":
            ffmpeg_path = shlex.quote(self.ffmpeg_path)
            video_path = shlex.quote(self.video_path)
            output_dir = shlex.quote(self.output_dir)

        crop = FramesExtractor.detect_crop(
            video_path=video_path, frames=3, ffmpeg_path=ffmpeg_path,
            video_length=video_length
        )

        command = [
            str(ffmpeg_path),
            "-i",
            str(video_path),
            *crop,
            "-s",
            "144x144",
            "-r",
            str(self.interval),
            str(output_dir)+"video_frame_%07d.jpeg",
        ]

        process = Popen(command, stdin=DEVNULL, stdout=PIPE, stderr=PIPE)
        output, error = process.communicate()

        ffmpeg_output = output.decode()
        ffmpeg_error = error.decode()

        if len(os.listdir(self.output_dir)) == 0:

            raise FFmpegFailedToExtractFrames(
                f"FFmpeg could not extract any frames.\n{command}\n{ffmpeg_output}\n{ffmpeg_error}"
            )



# This module has some code from <https://github.com/samdobson/image_slicer>

# Copyright (c) 2013 Sam Dobson

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
from math import ceil, floor, sqrt

from PIL import Image

from .utils import get_list_of_all_files_in_dir


class Tile:
    """Represents a single tile."""

    def __init__(self, image, number, position, coords, filename=None):
        self.image = image
        self.number = number
        self.position = position
        self.coords = coords
        self.filename = filename

    @property
    def row(self):
        return self.position[0]

    @property
    def column(self):
        return self.position[1]

    def generate_filename(
        self, directory=os.getcwd(), prefix="tile", file_format="png", path=True
    ):
        """Construct and return a filename for this tile."""
        filename = prefix + "_{col:02d}_{row:02d}.{ext}".format(
            col=self.column,
            row=self.row,
            ext=file_format.lower().replace("jpeg", "jpg"),
        )
        if not path:
            return filename
        return os.path.join(directory, filename)

    def save(self, filename=None, file_format="png"):
        if not filename:
            filename = self.generate_filename(file_format=file_format)
        self.image.save(filename, file_format)
        self.filename = filename


def calc_columns_rows(n):
    """
    Calculate the number of columns and rows required to divide an image
    into ``n`` parts.
    Return a tuple of integers in the format (num_columns, num_rows)
    """
    num_columns = int(ceil(sqrt(n)))
    num_rows = int(ceil(n / float(num_columns)))
    return (num_columns, num_rows)


def validate_image(number_tiles):
    """Basic sanity checks prior to performing a split."""
    TILE_LIMIT = 99 * 99

    try:
        number_tiles = int(number_tiles)
    except BaseException:
        raise ValueError("number_tiles could not be cast to integer.")

    if number_tiles > TILE_LIMIT or number_tiles < 2:
        raise ValueError(
            "Number of tiles must be between 2 and {} (you \
                          asked for {}).".format(
                TILE_LIMIT, number_tiles
            )
        )


def save_tiles(tiles, prefix="", directory=os.getcwd(), file_format="png"):
    """
    Write image files to disk. Create specified folder(s) if they
       don't exist. Return list of :class:`Tile` instance.
    Args:
       tiles (list):  List, tuple or set of :class:`Tile` objects to save.
       prefix (str):  Filename prefix of saved tiles.
    Kwargs:
       directory (str):  Directory to save tiles. Created if non-existant.
    Returns:
        Tuple of :class:`Tile` instances.
    """
    for tile in tiles:
        tile.save(
            filename=tile.generate_filename(
                prefix=prefix, directory=directory, file_format=file_format
            ),
            file_format=file_format,
        )
    return tuple(tiles)


def slicer(
    filename,
    number_tiles=None,
    col=None,
    row=None,
):

    Image.MAX_IMAGE_PIXELS = None

    im = Image.open(filename)
    im_w, im_h = im.size

    columns = 0
    rows = 0
    if number_tiles:
        validate_image(number_tiles)
        columns, rows = calc_columns_rows(number_tiles)

    tile_w, tile_h = int(floor(im_w / columns)), int(floor(im_h / rows))

    tiles = []
    number = 1
    for pos_y in range(0, im_h - rows, tile_h):  # -rows for rounding error.
        for pos_x in range(0, im_w - columns, tile_w):  # as above.
            area = (pos_x, pos_y, pos_x + tile_w, pos_y + tile_h)
            image = im.crop(area)
            position = (int(floor(pos_x / tile_w)) + 1, int(floor(pos_y / tile_h)) + 1)
            coords = (pos_x, pos_y)
            tile = Tile(image, number, position, coords)
            tiles.append(tile)
            number += 1
    im.close()
    return tuple(tiles)


def concatenate_video_frames_horizontally(
    frames_dir, horizontally_concatenated_image_path
) -> None:
    image_file_names = get_list_of_all_files_in_dir(frames_dir)
    total_images = len(image_file_names)
    first_image_filename = image_file_names[0]
    with Image.open(first_image_filename) as first_frame_image_in_list:
        width, height = first_frame_image_in_list.size

    base_image = Image.new("RGB", (width * total_images, height))

    x_offset = 0
    for image_filename in image_file_names:
        img = Image.open(image_filename)
        base_image.paste(img, (x_offset, 0))
        img.close()
        x_offset += width

    base_image.save(horizontally_concatenated_image_path)


def make_tile(frames_dir, horizontally_concatenated_image_path, tiles_dir) -> None:
    concatenate_video_frames_horizontally(
        frames_dir, horizontally_concatenated_image_path
    )
    tiles = list(
        slicer(
            horizontally_concatenated_image_path,
            number_tiles=64,
            col=8,
            row=8,
        )
    )
    save_tiles(tiles, prefix="tile", directory=tiles_dir, file_format="png")



from shutil import which
from subprocess import PIPE, Popen

from .exceptions import DownloadFailed, DownloadOutPutDirDoesNotExist
from .utils import does_path_exists, get_list_of_all_files_in_dir

# Python module to download the video from the input URL.
# Uses yt-dlp to download the video.


class Download:
    """
    Class that downloads the video prior to frames extraction.

    Tries to download the lowest quality video possible.
    Uses yt-dlp to download the videos.
    """

    def __init__(
        self,
        url: str,
        output_dir: str,
        worst: bool = True,
    ) -> None:
        """
        :param url: The URL of the video. The video will be
                    downloaded from this url. Must be a string.

        :param output_dir: The directory where the downloaded video will be stored.
                           Must be a string and path must be absolute.

        :param worst: The quality of video downloaded by yt-dlp.
                      True for worst quality and False for the default settings
                      of the downloader. Default value for worst is True.

        :return: None

        :rtype: NoneType
        """
        self.url = url
        self.output_dir = output_dir
        self.worst = worst

        if not does_path_exists(self.output_dir):
            raise DownloadOutPutDirDoesNotExist(
                f"No directory found at '{self.output_dir}' for storing the downloaded video. Can not download the video."
            )

        self.yt_dlp_path = str(which("yt-dlp"))
        self.download_video()

    def download_video(self) -> None:
        """Download the video from URL

        :return: None

        :rtype: NoneType

        """
        worst = " "
        if self.worst:
            worst = " -f worst "

        command = (
            f'"{self.yt_dlp_path}"'
            + worst
            + " "
            + '"'
            + self.url
            + '"'
            + " -o "
            + '"'
            + self.output_dir
            + "video_file.%(ext)s"
            + '"'
        )

        process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
        output, error = process.communicate()
        yt_dlp_output = output.decode()
        yt_dlp_error = error.decode()

        if len(get_list_of_all_files_in_dir(self.output_dir)) == 0:
            raise DownloadFailed(
                f"'{self.yt_dlp_path}' failed to download the video at"
                + f" '{self.url}'.\n{yt_dlp_output}\n{yt_dlp_error}"
            )



from .videohash import VideoHash
import time

urls = ["https://t.me/bloodysx/29195",
        "https://t.me/bloodysx/28649",
        "https://t.me/bloodysx/28589",
        "https://t.me/bloodysx/28571",
        "https://t.me/bloodysx/28408",
        "https://t.me/bloodysx/27870",
        "https://t.me/bloodysx/27321",
        "https://t.me/bloodysx/27261",
        "https://t.me/bloodysx/29211"
        ]

start = time.time()

for url in urls:
    VideoHash(url=url, frame_interval=12)

end = time.time() - start
print(f"Completed in {end} seconds.")


import os
import random
import re
import shutil
from pathlib import Path
from math import ceil
from typing import List, Optional, Union

import imagehash
import numpy as np
from imagedominantcolour import DominantColour
from PIL import Image

from .collagemaker import MakeCollage
from .downloader import Download
from .exceptions import DidNotSupplyPathOrUrl, StoragePathDoesNotExist
from .framesextractor import FramesExtractor
from .tilemaker import make_tile
from .utils import (
    create_and_return_temporary_directory,
    does_path_exists,
    get_list_of_all_files_in_dir,
)
from .videoduration import video_duration


class VideoHash:
    """
    VideoHash class provides an interface for computing & comparing the video
    hash values for videos(codec, containers etc) supported by FFmpeg.
    """

    def __init__(
        self,
        path: Optional[str] = None,
        url: Optional[str] = None,
        storage_path: Optional[str] = None,
        download_worst: bool = False,
        frame_interval: Union[int, float] = 1,
        do_not_copy: Optional[bool] = True,
    ) -> None:
        """
        :param path: Absolute path of the input video file.

        :param url: URL of the input video file. Every URL that is supported by
                    the yt-dlp package can be passed.

        :param storage_path: Storage path for the files created/downloaded by
                             the instance, pass the absolute path of the
                             directory.
                             If no argument is passed then the instance will
                             itself create the storage directory inside the
                             temporary directory of the system.

        :param download_worst: If set to True, download the worst quality video.
                               The default value is False, set True to conserve bandwidth.

        :param frame_interval: Number of frames extracted per unit time, the
                               default value is 1 per unit time. For 1 frame
                               per 5 seconds pass 1/5 or 0.2. For 5 fps pass 5.
                               Smaller frame_interval implies fewer frames and
                               vice-versa.


        :return: None

        :rtype: NoneType
        """
        self.path = path
        self.url = url

        self.storage_path = ""
        if storage_path:
            self.storage_path = storage_path

        self._storage_path = self.storage_path
        self.download_worst = download_worst
        self.do_not_copy = do_not_copy
        self.frame_interval = frame_interval

        self.task_uid = VideoHash._get_task_uid()

        self._create_required_dirs_and_check_for_errors()

        self._copy_video_to_video_dir()

        self.video_duration = video_duration(self.video_path)

        FramesExtractor(
            self.video_path,
            self.frames_dir,
            video_length=self.video_duration,
            interval=self.frame_interval,
        )

        self.collage_path = os.path.join(self.collage_dir, "collage.jpg")

        self.horizontally_concatenated_image_path = os.path.join(
            self.horizontally_concatenated_image_dir,
            "horizontally_concatenated_image.png",
        )

        MakeCollage(
            get_list_of_all_files_in_dir(self.frames_dir),
            self.collage_path,
            collage_image_width=1024,
        )

        make_tile(
            self.frames_dir, self.horizontally_concatenated_image_path, self.tiles_dir
        )

        self.image = Image.open(self.collage_path)
        self.bits_in_hash = 64
        self.similar_percentage = 15

        self._calc_hash()

    def __str__(self) -> str:
        """
        The video hash value of the instance. The hash value is 64 bit string
        prefixed with '0b', indicating the that the hash value is a bitstring.

        :return: The string representation of the instance. The video hash value
                 itself is the returned value.

        :rtype: str
        """

        return self.hash

    def __repr__(self) -> str:
        """
        Developer's representation of the VideoHash object.

        :return: Developer's representation of the instance.

        :rtype: str
        """

        return (
            f"VideoHash(hash={self.hash}, hash_hex={self.hash_hex}, "
            + f"collage_path={self.collage_path}, bits_in_hash={self.bits_in_hash})"
        )

    def __len__(self) -> int:
        """
        Length of the hash value string. Total length is 66 characters, 64 for
        the bitstring and 2 for the prefix '0b'.

        :return: Length of the the hash value, including the prefix '0b'.

        :rtype: int
        """
        return len(self.hash)

    def __ne__(self, other: object) -> bool:
        """
        Definition of the "!=" operator for the VideoHash objects.

        Instance of the VideoHash class, lists(bitlist) and string prefixed with
         '0x' and '0b' are accepted other types.


        If the hamming distance of this instance and the other instance
        is zero returns False else returns True.

        :return: True if other object and instance do not have the same hash
                 value else False.

        :rtype: bool
        """

        if self == other:
            return False
        return True

    def __eq__(self, other: object) -> bool:
        """
        Definition of the '=' operator on VideoHash objects.

        Instance of the VideoHash class, lists(bitlist) and string prefixed with
         '0x' and '0b' are accepted other types.

        :return: True if other object and instance have the same hash
                 value else False.

        :rtype: bool
        """

        if self - other == 0:
            return True
        return False

    def __sub__(self, other: object) -> int:
        """
        Definition of the '-' operator on VideoHash objects.

        Instance of the VideoHash class, lists(bitlist) and string prefixed with
         '0x' and '0b' are accepted other types.

        The method checks that the binary strings are prefixed with '0b',
        hexadecimal strings prefixed with '0x' and if the string is not
        prefixed then raise ValueError.

        :return: The hamming distance of hash values of the instance and other
                 object.

        :rtype: int

        :raises TypeError: If the object passed is not an instance of string, list
        or VideoHash.

        :raises ValueError: If the length of the hash values/bitlist of the
                            instance and the other object are not equal.

        :raises TypeError: If the hash values are python string objects but do
                           not have '0x' or '0b' as prefix.
        """
        if other is None:
            raise TypeError("Other hash is None. And it should not be None.")

        if isinstance(other, str):

            if other.lower().startswith("0x"):

                return self.hamming_distance(
                    string_a=self.hash,
                    string_b=VideoHash.hex2bin(other.lower(), self.bits_in_hash),
                )

            elif other.lower().startswith("0b"):

                if len(other) != len(self.hash):
                    raise ValueError(
                        "Can not compare different bits hashes. You must supply a %d bits hash."
                        % self.bits_in_hash
                    )
                return self.hamming_distance(string_a=self.hash, string_b=other.lower())

            else:

                raise TypeError(
                    "Hash string must start with either '0x' for hexadecimal or '0b' for binary."
                )

        if isinstance(other, list):

            if len(other) != self.bits_in_hash:
                raise ValueError(
                    f"The list does not have {self.bits_in_hash} bits. Can not calculate hamming distance."
                )

            return self.hamming_distance(bitlist_a=self.bitlist, bitlist_b=other)

        if isinstance(other, VideoHash):
            return self.hamming_distance(
                bitlist_a=self.bitlist, bitlist_b=other.bitlist
            )

        raise TypeError(
            "To calculate difference both of the hashes must be either "
            + "hexadecimal/binary strings or instance of VideoHash class."
        )

    def _copy_video_to_video_dir(self) -> None:
        """
        Copy the video from the path to the video directory.

        Copying avoids issues such as the user or some other
        process deleting the instance files while we are still
        processing.

        If instead of the path the uploader specified an url,
        then download the video and copy the file to video
        directory.


        :return: None

        :rtype: NoneType

        :raises ValueError: If the path supplied by the end user
                            lacks an extension. E.g. webm, mkv and mp4.
        """
        self.video_path: str = ""

        if self.path:
            # create a copy of the video at self.storage_path
            match = re.search(r"\.([^.]+$)", self.path)

            if match:
                extension = match.group(1)

            else:
                raise ValueError("File name (path) does not have an extension.")

            self.video_path = os.path.join(self.video_dir, (f"video.{extension}"))

            if self.do_not_copy:
                os.symlink(self.path, self.video_path)
            else:
                shutil.copyfile(self.path, self.video_path)

        if self.url:

            Download(
                self.url,
                self.video_download_dir,
                worst=self.download_worst,
            )

            downloaded_file = get_list_of_all_files_in_dir(self.video_download_dir)[0]
            match = re.search(r"\.(.*?)$", downloaded_file)

            extension = "mkv"

            if match:
                extension = match.group(1)

            self.video_path = f"{self.video_dir}video.{extension}"

            if self.do_not_copy:
                os.symlink(downloaded_file, self.video_path)
            else:
                shutil.copyfile(downloaded_file, self.video_path)

    def _create_required_dirs_and_check_for_errors(self) -> None:
        """
        Creates important directories before the main processing starts.

        The instance files are stored in these directories, no need to worry
        about the end user or some other processes interfering with the instance
        generated files.


        :raises DidNotSupplyPathOrUrl: If the user forgot to specify both the
                                       path and the url. One of them must be
                                       specified for creating the object.

        :raises ValueError: If user passed both path and url. Only pass
                            one of them if the file is available on both
                            then pass the path only.

        :raises StoragePathDoesNotExist: If the storage path specified by the
                                         user does not exist.

        :return: None

        :rtype: NoneType
        """
        if not self.path and not self.url:
            raise DidNotSupplyPathOrUrl(
                "You must specify either a path or an URL of the video."
            )

        if self.path and self.url:
            raise ValueError("Specify either a path or an URL and NOT both.")

        if not self.storage_path:
            self.storage_path = create_and_return_temporary_directory()
        if not does_path_exists(self.storage_path):
            raise StoragePathDoesNotExist(
                f"Storage path '{self.storage_path}' does not exist."
            )

        os_path_sep = os.path.sep

        self.storage_path = os.path.join(
            self.storage_path, (f"{self.task_uid}{os_path_sep}")
        )

        self.video_dir = os.path.join(self.storage_path, (f"video{os_path_sep}"))
        Path(self.video_dir).mkdir(parents=True, exist_ok=True)

        self.video_download_dir = os.path.join(
            self.storage_path, (f"downloadedvideo{os_path_sep}")
        )
        Path(self.video_download_dir).mkdir(parents=True, exist_ok=True)

        self.frames_dir = os.path.join(self.storage_path, (f"frames{os_path_sep}"))
        Path(self.frames_dir).mkdir(parents=True, exist_ok=True)

        self.tiles_dir = os.path.join(self.storage_path, (f"tiles{os_path_sep}"))
        Path(self.tiles_dir).mkdir(parents=True, exist_ok=True)

        self.collage_dir = os.path.join(self.storage_path, (f"collage{os_path_sep}"))
        Path(self.collage_dir).mkdir(parents=True, exist_ok=True)

        self.horizontally_concatenated_image_dir = os.path.join(
            self.storage_path, (f"horizontally_concatenated_image{os_path_sep}")
        )
        Path(self.horizontally_concatenated_image_dir).mkdir(
            parents=True, exist_ok=True
        )

    def is_similar(self, other: object) -> bool:
        """
        If 'similar_percentage' of bits are similar
        the similar else not.
        """
        if self - other <= ceil((self.similar_percentage / 100) * self.bits_in_hash):
            return True
        else:
            return False

    def is_diffrent(self, other: object) -> bool:
        """
        Refer to the is_similar,
        if not similar then diffrent.
        """

        if not self.is_similar(other):
            return True
        else:
            return False

    def delete_storage_path(self) -> None:
        """
        Delete the storage_path directory tree.

        Remember that deleting the storage directory will also delete the
        collage, extracted frames, and the downloaded video. If you passed an
        argument to the storage_path that directory will not be deleted but
        only the files and directories created inside that directory by the
        instance will be deleted, this is a feature(not a bug) to ensure that
        multiple instances of the same program are not deleting the storage
        path while other instances still require that storage directory.

        Many OS delete the temporary directory on boot or they never delete it.
        If you will be calculating videohash-value for many videos and don't
        want to run out of storage don't forget to delete the storage path.

        :return: None

        :rtype: NoneType
        """
        directory = self.storage_path

        if not self._storage_path:
            directory = (
                os.path.dirname(os.path.dirname(os.path.dirname(self.storage_path)))
                + os.path.sep
            )

        shutil.rmtree(directory, ignore_errors=True, onerror=None)

    @staticmethod
    def _get_task_uid() -> str:
        """
        Returns an unique task id for the instance. Task id is used to
        differentiate the instance files from the other unrelated files.

        We want to make sure that only the instance is manipulating the instance files
        and no other process nor user by accident deletes or edits instance files while
        we are still processing.

        :return: instance's unique task id.

        :rtype: str
        """
        sys_random = random.SystemRandom()

        return "".join(
            sys_random.choice(
                "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
            )
            for _ in range(20)
        )

    def hamming_distance(
        self,
        string_a: Optional[str] = None,
        string_b: Optional[str] = None,
        bitlist_a: Optional[List[int]] = None,
        bitlist_b: Optional[List[int]] = None,
    ) -> int:
        """
        Computes the hamming distance of the input bitstrings or bitlists.
        string_a and string_b must be bitstrings.
        bitlist_a and bitlist_b must be python strings containing only the
        bits and not the prefix "0b".

        :return: Hamming distance of the input bitstrings or bitlists.

        :rtype: int

        :raises ValueError: The input strings are of unequal length or bitlists
                            have different number of bits. Hamming distance is
                            not defined for unequal length strings.
        """
        if bitlist_a and bitlist_b:

            if len(bitlist_a) != len(bitlist_b):
                raise ValueError(
                    "Bit lists have unequal number of bits."
                    + " Can not compute hamming distance. Hamming distance is undefined."
                )

            _bitlist_a = bitlist_a
            _bitlist_b = bitlist_b

        if string_a and string_b:

            if len(string_a) != len(string_b):
                raise ValueError(
                    "Strings are of unequal length. Can not compute hamming distance. Hamming distance is undefined."
                )

            if string_a == self.hash:
                _bitlist_a = self.bitlist

            else:
                _bitlist_a = list(map(int, string_a.replace("0b", "")))

            if string_b == self.hash:
                _bitlist_b = self.bitlist

            else:
                _bitlist_b = list(map(int, string_b.replace("0b", "")))

        return len(
            np.bitwise_xor(
                _bitlist_a,
                _bitlist_b,
            ).nonzero()[0]
        )

    @staticmethod
    def hex2bin(hexstr: str, padding: int) -> str:
        """
        Convert hexadecimal('0x' prefixed) to bitstring string prefixed
        with '0b'.

        :return: Padded binary representation(bitstring) of the hexadecimal
                 input.

        :rtype: str

        :raises ValueError: If input hexadecimal string is not prefixed
                            with '0x'.
        """
        if not hexstr.lower().startswith("0x"):
            raise ValueError("Input hexadecimal string must have '0x' as the prefix.")

        return "0b" + str(bin(int(hexstr.lower(), 0))).replace("0b", "").zfill(padding)

    @staticmethod
    def bin2hex(binstr: str) -> str:
        """
        Convert bitstring('0b' prefixed) to hexadecimal string prefixed
        with '0x'.

        :return: Hex representation of the input bitstring.

        :rtype: str

        :raises ValueError: It input binary string is not prefixed with '0b'.
        """

        if not binstr.lower().startswith("0b"):
            raise ValueError("Binary string must be prefixed with '0b'.")

        return str(hex(int(binstr, 2)))

    def _calc_hash(self) -> None:
        """
        Calculates the hash value by calling the whash(wavelet hash) method of
        imagehash package. The wavelet hash of the collage is the videohash for
        the original input video.

        End-user is not provided any access to the imagehash instance but
        instead the binary and hexadecimal equivalent of the result of
        wavelet-hash.

        :return: None

        :rtype: NoneType
        """

        self.bitlist: List = []

        self.whash_bitlist: List = []

        self.dominant_colour_bitlist: List = []

        for row in imagehash.whash(self.image).hash.astype(int).tolist():
            self.whash_bitlist.extend(row)

        dominant_colour_list = []
        for file_path in get_list_of_all_files_in_dir(self.tiles_dir):
            dominantcolour = DominantColour(file_path)
            dominant_colour_list.append(dominantcolour.dominant_colour)

        pixels = [
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "r",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "g",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "b",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
            "l",
        ]

        for i in range(64):
            bit = 0
            if pixels[i] == dominant_colour_list[i]:
                bit = 1

            self.dominant_colour_bitlist.append(bit)

        for i in range(64):
            x = self.dominant_colour_bitlist[i]
            y = self.whash_bitlist[i]

            if x == y and x == 1:
                self.bitlist.append(0)
            elif x == y and x == 0:
                self.bitlist.append(0)
            else:
                self.bitlist.append(1)

        self.hash: str = ""

        for bit in self.bitlist:

            if bit:
                self.hash += "1"
            else:
                self.hash += "0"

        # the binary value is prefixed with 0b.
        self.hash = f"0b{self.hash}"
        self.hash_hex: str = VideoHash.bin2hex(self.hash)



import os
from math import ceil, sqrt
from typing import List

from PIL import Image

from .exceptions import CollageOfZeroFramesError
from .utils import does_path_exists

# Module to create collage from list of images, the
# images are the extracted frames of the input video.


class MakeCollage:
    """
    Class that creates the collage from list of images.

    Collage that should be as close to the shape of a square.

    The images are arranged by timestamp of the frames, their
    index in the image_list is based on thier timestamp on the
    video. The image with the index 2 is a frame from the 3rd
    second and an index 39 is from at the 40th second. The index
    is one less due to zero-based indexing.


    Let's say we have a list with 9 images.

    As the images should be arranged in a way to resemble a
    square, we take the square root of 9 and that is 3. Now
    we need to make a 3x3 frames collage.

    Arrangement should be:
    Img1 Img2 Img3
    Img4 Img5 Img6
    Img7 Img8 Img9

    If the number of images is not a perfect square, calculate the
    square root and round it to the nearest integer.

    If number of images is 13, which is not a perfect square.

    sqrt(13) = 3.605551275463989
    round(3.605551275463989) = 4

    Thus the image should be 4x4 frames of collage.

    Arrangement should be:
    -----------------------------
    |  Img1  Img2  Img3   Img4  |
    |  Img5  Img6  Img7   Img8  |
    |  Img9  Img10 Img11  Img12 |
    |  Img13  X     X      X    |
    -----------------------------

    X denotes the empty space due to lack of images.
    But the empty spaces will not affect the robustness
    as downsized/transcoded version of the video will also
    produce these vacant spaces.
    """

    def __init__(
        self,
        image_list: List[str],
        output_path: str,
        collage_image_width: int = 1024,
    ) -> None:
        """
        Checks if the list passed is not an empty list.
        Also makes sure that the output_path directory exists.

        And calls the make method, the make method creates the collage.

        :param image_list: A python list containing the list of absolute
                           path of images that are to be added in the collage.
                           The order of images is kept intact and is very important.

        :param output_path: Absolute path of the collage including
                            the image name. (This is where the collage is saved.)
                            Example: '/home/username/projects/collage.jpeg'.

        :param collage_image_width: An integer specifying the image width of the
                                    output collage. Default value is 1024 pixels.

        :return: None

        :rtype: NoneType
        """
        self.image_list = image_list
        self.number_of_images = len(self.image_list)
        self.output_path = output_path
        self.collage_image_width = collage_image_width

        self.images_per_row_in_collage = int(round(sqrt(self.number_of_images)))

        if self.number_of_images == 0:
            raise CollageOfZeroFramesError("Can not make a collage of zero images.")

        output_path_dir = os.path.dirname(self.output_path) + "/"
        if not does_path_exists(output_path_dir):
            raise FileNotFoundError(
                "Directory at which output collage is to be saved does not exists."
            )

        self.make()

    def make(self) -> None:
        """
        Creates the collage from the list of images.

        It calculates the scale of the images on collage by
        measuring the first image width and height, there's no
        reason for choosing first one and it's arbitrary. But
        we assume that all the images passed should have same size.

        A base image of 'collage_image_width' width and of 'number
        of rows times scaled frame image height' height is created.
        The base image has all pixels with RGB value 0,0,0 that is
        the base image is pure black. The frame images are now embeded
        on it.
        The frame images are scaled to fit the collage base image such
        that the shape of collage is as close to the shape of a square.

        :return: None

        :rtype: NoneType
        """

        # arbitrarily selecting the first image from the list, index 0
        with Image.open(self.image_list[0]) as first_frame_image_in_list:

            # Find the width and height of the first image of the list.
            # Assuming all the images have same size.
            frame_image_width, frame_image_height = first_frame_image_in_list.size

        # scale is the ratio of collage_image_width and product of
        # images_per_row_in_collage with frame_image_width.

        # The scale will always lie between 0 and 1, which implies that
        # the images are always going to get downsized.
        scale = (self.collage_image_width) / (
            self.images_per_row_in_collage * frame_image_width
        )

        # Calculating the scaled height and width for the frame image.
        scaled_frame_image_width = ceil(frame_image_width * scale)
        scaled_frame_image_height = ceil(frame_image_height * scale)

        # Divide the number of images by images_per_row_in_collage. The later
        # was calculated by taking the square root of total number of images.
        number_of_rows = ceil(self.number_of_images / self.images_per_row_in_collage)

        # Multiplying the height of one downsized image with number of rows.
        # Height of 1 downsized image is product of scale and frame_image_height
        # Total height is number of rows times the height of one downsized image.
        self.collage_image_height = ceil(scale * frame_image_height * number_of_rows)

        # Create an image of passed collage_image_width and calculated collage_image_height.
        # The downsized images will be pasted on this new base image.
        # The image is 0,0,0 RGB(black).
        collage_image = Image.new(
            "RGB", (self.collage_image_width, self.collage_image_height)
        )

        # keep track of the x and y coordinates of the resized frame images
        i, j = (0, 0)

        # iterate the frames and paste them on their position on the collage_image
        for count, frame_path in enumerate(self.image_list):

            # Set the x coordinate to zero if we are on the first column
            # If self.images_per_row_in_collage is 4
            # then 0,4,8 and so on should have their x coordinate as 0
            if (count % self.images_per_row_in_collage) == 0:
                i = 0

            # open the frame image, must open it to resize it using the thumbnail method
            frame = Image.open(frame_path)

            # scale the opened frame images
            frame.thumbnail(
                (scaled_frame_image_width, scaled_frame_image_height),
                Image.Resampling.LANCZOS,
            )

            # set the value of x to that of i's value.
            # i is set to 0 if we are on the first column.
            x = i

            # It ensures that y coordinate stays the same for any given row.
            # The floor of a real number is the largest integer that is less
            # than or equal to the number. floor division is used because of
            # the zero based indexing, the floor of the division stays same
            # for an entier row as the decimal values are negled by the floor.
            # for the first row the result of floor division is always zero and
            # the product of 0 with scaled_frame_image_height is also zero, they
            # y coordinate for the first row is 0.
            # For the second row the result of floor division is one and the prodcut
            # with scaled_frame_image_height ensures that the y coordinate is
            # scaled_frame_image_height below the first row.
            y = (j // self.images_per_row_in_collage) * scaled_frame_image_height

            # paste the frame image on the newly created base image(base image is black)
            collage_image.paste(frame, (x, y))
            frame.close()

            # increase the x coordinate by scaled_frame_image_width
            # to get the x coordinate of the next frame. unless the next image
            # will be on the very first column this will be the x coordinate.
            i = i + scaled_frame_image_width

            # increase the value of j by 1, this is to calculate the y coordinate of
            # next image. The increased number will be floor divided by images_per_row_in_collage
            # therefore the y coordinate stays the same for any given row.
            j += 1

        # save the base image with all the scaled frame images embeded on it.
        collage_image.save(self.output_path)
        collage_image.close()



import os
import tempfile
from pathlib import Path
from typing import List


def get_list_of_all_files_in_dir(directory: str) -> List[str]:
    """
    Returns a list containing all the file paths(absolute path) in a directory.
    The list is sorted.

    :return: List of absolute path of all files in a directory.

    :rtype: List[str]
    """
    return sorted([(directory + filename) for filename in os.listdir(directory)])


def does_path_exists(path: str) -> bool:
    """
    If a directory is supplied then check if it exists.
    If a file is supplied then check if it exists.

    If directory/file exists returns True else returns False

    :return: True if dir or file exists else False.

    :rtype: bool
    """
    if os.path.isdir(path) or os.path.isfile(path):
        return os.path.exists(path)

    else:
        # it's file
        return False


def create_and_return_temporary_directory() -> str:
    """
    create a temporary directory where we can store the video, frames and the
    collage.

    :return: Absolute path of the empty directory.

    :rtype: str
    """
    path = os.path.join(tempfile.mkdtemp(), ("temp_storage_dir" + os.path.sep))
    Path(path).mkdir(parents=True, exist_ok=True)
    return path



__title__ = "videohash2"
__description__ = (
    "Python package for Near Duplicate Video Detection "
    + "(Perceptual Video Hashing) - Get a 64-bit comparable "
    + "hash-value for any video."
)

__url__ = "https://demmenie.github.io/videohash2/"
__version__ = "3.1.0"
__status__ = "production"
__author__ = "Akash Mahanty and Chico Demmenie"
__author_email__ = "cdemmenie@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright 2024 Akash Mahanty"



import re
from shutil import which
from subprocess import PIPE, Popen
from typing import Optional

# Module to determine the length of video.
# The length is found by the FFmpeg, the output of video_duration is in seconds.


def video_duration(video_path: str, ffmpeg_path: Optional[str] = None) -> float:
    """
    Retrieve the exact video duration as echoed by the FFmpeg and return
    the duration in seconds. Maximum duration supported is 999 hours, above
    which the regex is doomed to fail(no match).

    :param video_path: Absolute path of the video file.

    :param ffmpeg_path: Path of the FFmpeg software if not in path.

    :return: Video length(duration) in seconds.

    :rtype: float
    """

    if not ffmpeg_path:
        ffmpeg_path = str(which("ffmpeg"))

    command = f'"{ffmpeg_path}" -i "{video_path}"'
    process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
    output, error = process.communicate()

    match = re.search(
        r"Duration\:(\s\d?\d\d\:\d\d\:\d\d\.\d\d)\,",
        (output.decode() + error.decode()),
    )

    if match:
        duration_string = match.group(1)

    hours, minutes, seconds = duration_string.strip().split(":")

    return float(hours) * 60.00 * 60.00 + float(minutes) * 60.00 + float(seconds)



import os.path

from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), "README.md"), encoding="utf-8") as f:
    long_description = f.read()

about = {}
with open(os.path.join(os.path.dirname(__file__), "videohash2", "__version__.py")) as f:
    exec(f.read(), about)

version = str(about["__version__"])

download_url = f"https://github.com/demmenie/videohash2/archive/{version}.tar.gz"

setup(
    name=about["__title__"],
    packages=["videohash2"],
    version=version,
    description=about["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=about["__license__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    download_url=download_url,
    keywords=[
        "videohash2",
        "near duplicate video detection",
        "NDVD",
        "perceptual video hashing",
        "video hashing",
        "near duplicate video",
        "compare videos",
        "video",
        "video diff",
    ],
    install_requires=[
        "Pillow",
        "ImageHash",
        "imagedominantcolour",
        "yt-dlp",
    ],
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Information Technology",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Operating System :: POSIX :: BSD",
        "Topic :: Multimedia :: Video",
        "Natural Language :: English",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Typing :: Typed",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",        
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    project_urls={
        "Source": "https://github.com/demmenie/videohash2",
        "Documentation": "https://github.com/akamhy/videohash/wiki",
        "Tracker": "https://github.com/demmenie/videohash2/issues",
    },
)



"""
videohash.exceptions
~~~~~~~~~~~~~~~~~~~
This module contains the videohash's exceptions.
"""


class VideoHashError(Exception):
    """Base Exception for the videohash package."""

    pass


class StoragePathDoesNotExist(VideoHashError):
    """
    The storage path passed by the user does not exist.
    The collage is the image representing your video as an two dimensional bitmap image.
    """

    pass


class FramesExtractorOutPutDirDoesNotExist(VideoHashError):
    """The frames output directory passed to the frame extractor does not exist."""

    pass


class DownloadOutPutDirDoesNotExist(VideoHashError):
    """The output directory passed to downloader for storing the downloaded video does not exist."""

    pass


class DidNotSupplyPathOrUrl(VideoHashError):
    """Must supply either a path for the video or a valid URL"""

    pass


class CollageOfZeroFramesError(VideoHashError):
    """Raised if zero frames are passed for collage making."""

    pass


class DownloadFailed(VideoHashError):
    """Download software failed to download the video from the url."""

    pass


class FFmpegError(VideoHashError):
    """Base error for the FFmpeg software."""

    pass


class FFmpegNotFound(FFmpegError):
    """FFmpeg is either not installed or not in the executable path of the system."""

    pass


class FFmpegFailedToExtractFrames(FFmpegError):
    """FFmpeg failed to extract any frame at all. Maybe the input video is damaged or corrupt."""

    pass



# ██╗░░░██╗██╗██████╗░███████╗░█████╗░██╗░░██╗░█████╗░░██████╗██╗░░██╗
# ██║░░░██║██║██╔══██╗██╔════╝██╔══██╗██║░░██║██╔══██╗██╔════╝██║░░██║
# ╚██╗░██╔╝██║██║░░██║█████╗░░██║░░██║███████║███████║╚█████╗░███████║
# ░╚████╔╝░██║██║░░██║██╔══╝░░██║░░██║██╔══██║██╔══██║░╚═══██╗██╔══██║
# ░░╚██╔╝░░██║██████╔╝███████╗╚█████╔╝██║░░██║██║░░██║██████╔╝██║░░██║
# ░░░╚═╝░░░╚═╝╚═════╝░╚══════╝░╚════╝░╚═╝░░╚═╝╚═╝░░╚═╝╚═════╝░╚═╝░░╚═╝


"""
The Python package for near duplicate video detection

https://github.com/akamhy/videohash

Usage:

>>> from videohash2 import VideoHash
>>> # video: Artemis I Hot Fire Test
>>> videohash1 = VideoHash(url="https://www.youtube.com/watch?v=PapBjpzRhnA", download_worst=False)
>>>
>>> videohash1.hash # video hash value of the file, value is same as str(videohash1)
'0b0011010000011111111011111111111110001111011110000000000000000000'
>>>
>>> #VIDEO:Artemis I Hot Fire Test
>>> url2="https://raw.githubusercontent.com/akamhy/videohash/main/assets/rocket.mkv"
>>> videohash2 = VideoHash(url=url2)
>>> videohash2.hash
'0b0011010000011111111011111111111110001111011110000000000000000000'
>>> videohash2.hash_hex
'0x341fefff8f780000'
>>> videohash2.hash_hex
'0x341fefff8f780000'
>>> videohash1 - videohash2
0
>>> videohash1 == videohash2
True
>>> videohash1 == "0b0011010000011111111011111111111110001111011110000000000000000000"
True
>>> videohash1 != videohash2
False
>>> path3 = "/home/akamhy/Downloads/rocket.mkv" #VIDEO: Artemis I Hot Fire Test
>>> videohash3 = VideoHash(path=path3)
>>> videohash3.hash
'0b0011010000011111111011111111111110001111011110000000000000000000'
>>> videohash3 - videohash2
0
>>> videohash3 == videohash1
True
>>> url4 = "https://www.youtube.com/watch?v=_T8cn2J13-4" #VIDEO: How We Are Going to the Moon
>>> videohash4 = VideoHash(url=url4)
>>> videohash4.hash_hex
'0x7cffff000000eff0'
>>> videohash4 - "0x7cffff000000eff0"
0
>>> videohash4.hash
'0b0111110011111111111111110000000000000000000000001110111111110000'
>>> videohash4 - videohash2
34
>>> videohash4 != videohash2
True

Run the above code @ https://replit.com/@akamhy/videohash-usage-2xx-example-code-for-video-hashing#main.py

Extended Usage : https://github.com/akamhy/videohash/wiki/Extended-Usage

API Reference : https://github.com/akamhy/videohash/wiki/API-Reference


:copyright: (c) 2021 Akash Mahanty
:license: MIT, see LICENSE for more details.
:pypi: https://pypi.org/project/videohash2/
:wiki: https://github.com/akamhy/videohash/wiki
:cite: https://doi.org/10.5281/zenodo.4448295
"""

from .__version__ import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __status__,
    __title__,
    __url__,
    __version__,
)
from .exceptions import (
    CollageOfZeroFramesError,
    DidNotSupplyPathOrUrl,
    DownloadFailed,
    DownloadOutPutDirDoesNotExist,
    FFmpegError,
    FFmpegFailedToExtractFrames,
    FFmpegNotFound,
    FramesExtractorOutPutDirDoesNotExist,
    StoragePathDoesNotExist,
    VideoHashError,
)
from .videoduration import video_duration
from .videohash import VideoHash
